from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
)
from fa_purity._bug import (
    LibraryBug,
)
from typing import (
    Callable,
    Generic,
    TypeVar,
    Union,
)

_L = TypeVar("_L")
_R = TypeVar("_R")
_T = TypeVar("_T")


@dataclass(frozen=True)
class _Empty:
    pass


@dataclass(frozen=True)
class _Coproduct(Generic[_L, _R]):
    left: Union[_L, _Empty]
    right: Union[_R, _Empty]
    left_val: bool

    @staticmethod
    def assert_non_empty(item: Union[_T, _Empty]) -> _T:
        if isinstance(item, _Empty):
            raise LibraryBug(
                Exception("assert_non_empty received empty input")
            )
        return item


@dataclass(frozen=True)
class Coproduct(Generic[_L, _R]):
    _inner: _Coproduct[_L, _R]

    @staticmethod
    def inl(value: _L) -> Coproduct[_L, _R]:
        return Coproduct(_Coproduct(value, _Empty(), True))

    @staticmethod
    def inr(value: _R) -> Coproduct[_L, _R]:
        return Coproduct(_Coproduct(_Empty(), value, False))

    def map(
        self, transform_1: Callable[[_L], _T], transform_2: Callable[[_R], _T]
    ) -> _T:
        if self._inner.left_val:
            return transform_1(_Coproduct.assert_non_empty(self._inner.left))
        return transform_2(_Coproduct.assert_non_empty(self._inner.right))


@dataclass(frozen=True)
class CoproductFactory(Generic[_L, _R]):
    def inl(self, value: _L) -> Coproduct[_L, _R]:
        return Coproduct.inl(value)

    def inr(self, value: _R) -> Coproduct[_L, _R]:
        return Coproduct.inr(value)



from deprecated import (
    deprecated,
)
from fa_purity._core.cmd import (
    Cmd,
    CmdUnwrapper,
    unsafe_unwrap,
)
from typing import (
    Callable,
    TypeVar,
)

_A = TypeVar("_A")


@deprecated("NEW API: use `Cmd.new_cmd` instead")  # type: ignore[misc]
def new_cmd(action: Callable[[CmdUnwrapper], _A]) -> Cmd[_A]:
    return Cmd.new_cmd(action)


__all__ = [
    "Cmd",
    "CmdUnwrapper",
    "unsafe_unwrap",
]



from deprecated import (
    deprecated,
)
from fa_purity._core.coproduct import (
    Coproduct,
    CoproductFactory,
    CoproductTransform,
    UnionFactory,
)
from typing import (
    Optional,
    Type,
    TypeVar,
    Union,
)

_L = TypeVar("_L")
_R = TypeVar("_R")


@deprecated("NEW API: use `UnionFactory` instead")  # type: ignore[misc]
def inr(val: _R, _left: Optional[Type[_L]] = None) -> Union[_L, _R]:
    return val


@deprecated("NEW API: use `UnionFactory` instead")  # type: ignore[misc]
def inl(val: _L, _right: Optional[Type[_R]] = None) -> Union[_L, _R]:
    return val


__all__ = [
    "Coproduct",
    "CoproductFactory",
    "CoproductTransform",
    "UnionFactory",
]



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
)
from fa_purity._core.coproduct import (
    Coproduct,
    CoproductFactory,
    CoproductTransform,
)
from typing import (
    Callable,
    Generic,
    Optional,
    Type,
    TypeVar,
    Union,
)

_S = TypeVar("_S")
_F = TypeVar("_F")
_T = TypeVar("_T")


@dataclass(frozen=True)
class _Result(Generic[_S, _F]):
    value: Coproduct[_S, _F]


@dataclass(frozen=True)
class PureResult(Generic[_S, _F]):
    "Equivalent to Coproduct[_S, _F], but designed to handle explicit errors"
    _inner: _Result[_S, _F]

    @staticmethod
    def success(
        val: _S, _type: Optional[Type[_F]] = None
    ) -> PureResult[_S, _F]:
        item: Coproduct[_S, _F] = Coproduct.inl(val)
        return PureResult(_Result(item))

    @staticmethod
    def failure(
        val: _F, _type: Optional[Type[_S]] = None
    ) -> PureResult[_S, _F]:
        item: Coproduct[_S, _F] = Coproduct.inr(val)
        return PureResult(_Result(item))

    def map(self, function: Callable[[_S], _T]) -> PureResult[_T, _F]:
        factory: CoproductFactory[_T, _F] = CoproductFactory()
        val = self._inner.value.map(
            lambda s: factory.inl(function(s)),
            lambda f: factory.inr(f),
        )
        return PureResult(_Result(val))

    def alt(self, function: Callable[[_F], _T]) -> PureResult[_S, _T]:
        factory: CoproductFactory[_S, _T] = CoproductFactory()
        val = self._inner.value.map(
            lambda s: factory.inl(s),
            lambda f: factory.inr(function(f)),
        )
        return PureResult(_Result(val))

    def to_coproduct(self) -> Coproduct[_S, _F]:
        return self._inner.value

    def bind(
        self, function: Callable[[_S], PureResult[_T, _F]]
    ) -> PureResult[_T, _F]:
        factory: CoproductFactory[_T, _F] = CoproductFactory()
        val = self._inner.value.map(
            lambda s: function(s).to_coproduct(),
            lambda f: factory.inr(f),
        )
        return PureResult(_Result(val))

    def lash(
        self, function: Callable[[_F], PureResult[_S, _T]]
    ) -> PureResult[_S, _T]:
        factory: CoproductFactory[_S, _T] = CoproductFactory()
        val = self._inner.value.map(
            lambda s: factory.inl(s),
            lambda f: function(f).to_coproduct(),
        )
        return PureResult(_Result(val))

    def swap(self) -> PureResult[_F, _S]:
        val = CoproductTransform(self._inner.value).swap()
        return PureResult(_Result(val))

    def apply(
        self, wrapped: PureResult[Callable[[_S], _T], _F]
    ) -> PureResult[_T, _F]:
        return wrapped.bind(lambda f: self.map(f))

    def cop_value_or(self, default: _T) -> Coproduct[_S, _T]:
        factory: CoproductFactory[_S, _T] = CoproductFactory()
        val = self._inner.value.map(
            lambda s: factory.inl(s),
            lambda _: factory.inr(default),
        )
        return val

    def cop_or_else_call(
        self, function: Callable[[], _T]
    ) -> Coproduct[_S, _T]:
        factory: CoproductFactory[_S, _T] = CoproductFactory()
        val = self._inner.value.map(
            lambda s: factory.inl(s),
            lambda _: factory.inr(function()),
        )
        return val

    def value_or(self, default: _T) -> Union[_S, _T]:
        return CoproductTransform(self.cop_value_or(default)).to_union()

    def or_else_call(self, function: Callable[[], _T]) -> Union[_S, _T]:
        return CoproductTransform(self.cop_or_else_call(function)).to_union()

    def to_union(self) -> Union[_S, _F]:
        return CoproductTransform(self._inner.value).to_union()



from ._core import (
    Coproduct,
    CoproductFactory,
)
from dataclasses import (
    dataclass,
)
from typing import (
    Generic,
    TypeVar,
    Union,
)

_A = TypeVar("_A")
_B = TypeVar("_B")
_C = TypeVar("_C")
_L = TypeVar("_L")
_R = TypeVar("_R")


@dataclass(frozen=True)
class UnionFactory(Generic[_L, _R]):
    def inl(self, value: _L) -> Union[_L, _R]:
        return value

    def inr(self, value: _R) -> Union[_L, _R]:
        return value


@dataclass(frozen=True)
class CoproductTransform(Generic[_L, _R]):
    _value: Coproduct[_L, _R]

    def swap(self) -> Coproduct[_R, _L]:
        def _right(item: _R) -> Coproduct[_R, _L]:
            return Coproduct.inl(item)

        def _left(item: _L) -> Coproduct[_R, _L]:
            return Coproduct.inr(item)

        return self._value.map(_left, _right)

    def to_union(self) -> Union[_L, _R]:
        factory: UnionFactory[_L, _R] = UnionFactory()
        return self._value.map(
            lambda l: factory.inl(l),
            lambda r: factory.inr(r),
        )

    @staticmethod
    def permutate(
        item: Coproduct[_A, Coproduct[_B, _C]]
    ) -> Coproduct[Coproduct[_A, _B], _C]:
        factory: CoproductFactory[Coproduct[_A, _B], _C] = CoproductFactory()
        return item.map(
            lambda a: factory.inl(Coproduct.inl(a)),
            lambda bc: bc.map(
                lambda b: factory.inl(Coproduct.inr(b)),
                lambda c: Coproduct.inr(c),
            ),
        )



from ._core import (
    Coproduct,
    CoproductFactory,
)
from ._transform import (
    CoproductTransform,
    UnionFactory,
)

__all__ = [
    "UnionFactory",
    "Coproduct",
    "CoproductFactory",
    "CoproductTransform",
]



from arch_lint.dag import (
    DagMap,
)
from arch_lint.graph import (
    FullPathModule,
)
from fa_purity import (
    FrozenList,
)
from typing import (
    Dict,
    FrozenSet,
    TypeVar,
    Union,
)

_dag: Dict[str, FrozenList[Union[FrozenList[str], str]]] = {
    "fa_purity": (
        "flatten",
        "stream",
        ("json_2", "date_time"),
        (
            "pure_iter",
            "json",
        ),
        "_iter_factory",
        "maybe",
        ("lock", "cmd", "result"),
        "frozen",
        "union",
        "_core",
        ("_bug", "utils"),
    ),
    "fa_purity._core": ("result", ("cmd", "coproduct")),
    "fa_purity._core.coproduct": ("_transform", "_core"),
    "fa_purity.stream": (
        "transform",
        "factory",
        "core",
        "_inner",
    ),
    "fa_purity.pure_iter": (
        "transform",
        "factory",
        "core",
        "_inner",
    ),
    "fa_purity.result": (
        "transform",
        "factory",
        "core",
    ),
    "fa_purity.cmd": (
        "transform",
        "core",
    ),
    "fa_purity.json_2": (
        "value",
        "primitive",
    ),
    "fa_purity.json_2.primitive": (
        ("_transform", "_factory"),
        "_core",
    ),
    "fa_purity.json_2.value": (
        ("_transform", "_factory"),
        "_core",
    ),
    "fa_purity.json": (
        "transform",
        "factory",
        "value",
        "primitive",
        "errors",
    ),
    "fa_purity.json.primitive": (
        "factory",
        "core",
    ),
    "fa_purity.json.value": (
        "transform",
        "factory",
        "core",
    ),
}
_T = TypeVar("_T")


def raise_or_return(item: Union[Exception, _T]) -> _T:
    if isinstance(item, Exception):
        raise item
    return item


def project_dag() -> DagMap:
    return raise_or_return(DagMap.new(_dag))


def forbidden_allowlist() -> Dict[FullPathModule, FrozenSet[FullPathModule]]:
    _raw: Dict[str, FrozenSet[str]] = {}
    return {
        raise_or_return(FullPathModule.from_raw(k)): frozenset(
            raise_or_return(FullPathModule.from_raw(i)) for i in v
        )
        for k, v in _raw.items()
    }



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
)
from fa_purity._core.cmd import (
    Cmd,
    CmdUnwrapper,
)
from threading import (
    Lock as _Lock,
)
from typing import (
    TypeVar,
)

_T = TypeVar("_T")


@dataclass(frozen=True)
class ThreadLock:
    _inner: _Lock

    @staticmethod
    def new() -> Cmd[ThreadLock]:
        return Cmd.from_cmd(lambda: ThreadLock(_Lock()))

    def execute_with_lock(self, cmd: Cmd[_T]) -> Cmd[_T]:
        def _action(unwrapper: CmdUnwrapper) -> _T:
            with self._inner:
                return unwrapper.act(cmd)

        return Cmd.new_cmd(_action)

    @property
    def acquire(self) -> Cmd[None]:
        return Cmd.from_cmd(lambda: self._inner.acquire()).map(lambda _: None)

    @property
    def try_acquire(self) -> Cmd[bool]:
        "return False if not acquired"
        return Cmd.from_cmd(lambda: self._inner.acquire(False))

    @property
    def release(self) -> Cmd[None]:
        return Cmd.from_cmd(lambda: self._inner.release()).map(lambda _: None)

    @property
    def locked(self) -> Cmd[bool]:
        return Cmd.from_cmd(lambda: self._inner.locked())
